// $Id: World.cc,v 1.54 1999/12/14 11:26:33 grumbel Exp $
//
// Pingus - A free Lemmings clone
// Copyright (C) 1999 Ingo Ruhnke <grumbel@gmx.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <vector>
#include <cstdio>
#include <ctime>
#include <algorithm>
#include <functional>

#include "PinguBmpMap.hh"
#include "PingusSpotMap.hh"
#include "PinguRandomMap.hh"

#include "PingusError.hh"
#include "algo.hh"
#include "PingusSound.hh"
#include "globals.hh"
#include "Background.hh"
#include "World.hh"
#include "traps/traps.hh"
#include "Result.hh"
#include "Liquid.hh"
#include "actions/pingu_actions.hh"
#include "actions/bomber.hh"
#include "entrances/entrances.hh"
#include "FVec.hh"

// Workaround for a VC bug
#ifdef WIN32
#  define for if(0);else for
#endif /* WIN32 */

//CL_ResourceManager* World::resource;

struct WorldObj_less : public binary_function<WorldObj*, WorldObj*, bool>
{
  bool operator() (WorldObj* a, WorldObj* b) const
    {
      return (*a) < (*b);
    }
};

World::World()
{
  released_pingus = 0;
  saved_pingus = 0;

  exit_world = true;
}

World::World(PLF* plf)
{
  //  Result r;
  init(plf);
}

World::~World()
{
  // Bug: This crashes, if pingus is exited
  delete colmap;
  delete map;
  delete bg;

  for(vector<Exit*>::size_type i = 0; i < exits.size(); ++i) 
    delete exits[i];
  
  for(vector<Entrance*>::size_type i = 0; i < entrance.size(); ++i) 
    delete entrance[i];
 
  for(vector<Trap*>::size_type i = 0; i < traps.size(); ++i)
    delete traps[i];

  for(vector<Hotspot*>::size_type i = 0; i < hotspot.size(); ++i)
    delete hotspot[i];
  
  for(vector<Liquid*>::size_type i = 0; i < liquid.size(); ++i) {
    delete liquid[i];
  }
}

// Merge the different layers on the screen together
void 
World::draw_offset(int x_of, int y_of, float s)
{
  bg->draw_offset(x_of, y_of, s);

  //  map->draw_offset(x_of, y_of, s);

  for(vector<WorldObj*>::size_type i=0; i < world_obj.size(); ++i)
    world_obj[i]->draw_offset(x_of, y_of, s);

  // BUG: This give a warning while compiling, because the function
  // isn't const, but setting it const give some other problems.
  // Anyone knows how to solve this?
  particle.draw_offset(x_of, y_of, s);
}

void 
World::let_move()
{
  if (!exit_world && (allowed_pingus == released_pingus || do_armageddon)
      && pingus_out == 0) {
    cout << "World finished" << endl;
    exit_world = true;
    exit_time = GameTime::get_time() + 75;
  }

  if (do_armageddon && armageddon_count != pingus.end()) {
    if (armageddon_count == 0) {
      armageddon_count = pingus.begin();
    } else {
      int i = 0;
      for(PinguIter pingu = armageddon_count; i < 5; ++pingu, ++i) {
	armageddon_count = pingu;	
	if (armageddon_count == pingus.end())
	  break;
	if ((*pingu)->get_status() != dead)
	  (*pingu)->set_action(new Bomber);
	else
	  --i;
      }
    }
  }
  
  // Create new pingus, if enough time is passed
  if (!do_armageddon) {
    for(vector<entrance_data>::size_type i=0; i < entrance.size(); ++i) {
      if (pingus.size() < allowed_pingus
	  && entrance[i]->pingu_ready()) {
	pingus.push_back(entrance[i]->get_pingu());
	++released_pingus;
      }
    }
  }
  
  // Let all pingus move and
  // Let the pingus catch each other and
  // Let the traps catch the pingus and
  // Let the exit catch the pingus
  pingus_out = pingus.size();
  saved_pingus = 0;

  for(PinguIter pingu = pingus.begin(); pingu != pingus.end(); ++pingu) {
    (*pingu)->let_move();
    
    if ((*pingu)->get_status() == dead
	|| (*pingu)->get_status() == exited) {
      if ((*pingu)->get_status() == exited)
	++saved_pingus;
      --pingus_out;
      continue;
    }
    
    if ((*pingu)->need_catch()) {
      for(PinguIter j1=pingus.begin(); j1 != pingus.end(); ++j1) {
	(*pingu)->catch_pingu(*j1);
      }
    }
    
    for(vector<trap_data>::size_type j2=0; j2 < traps.size(); ++j2) 
      traps[j2]->catch_pingu(*pingu);
      
    for(vector<exit_data>::size_type j3=0; j3 < exits.size(); ++j3) 
      exits[j3]->catch_pingu(*pingu);
      
  }
  for(vector<trap_data>::size_type i=0; i < traps.size(); ++i) 
    traps[i]->let_move();

  for(vector<entrance_data>::size_type i2=0; i2 < entrance.size(); ++i2) 
    entrance[i2]->let_move();

  particle.let_move();
  bg->let_move();

  if (force_vectors)
    // Clear the explosion force list
    ForcesHolder::clear_explo_list();
}

void 
World::save()
{
  // Empty should save the state of the wolrd 
  // Maybe that should be implemented using a demo file
}

void 
World::load()
{
  // Empty should load the state of the world
  // Maybe that should be implemented using a demo file
}

void 
World::init(PLF* level_data)
{
  do_armageddon = false;
  armageddon_count = 0;

  exit_time = level_data->get_time();

  if (exit_time) {
    exit_world = true;
  } else {
    exit_world = false;
  }

  if (verbose)
    cout << "World init" << endl;

  // load the foreground map
  switch (level_data->map_type()) {
  case BMP:
    map = new PinguBmpMap(level_data->get_fg());
    break;
  case SPOT:
    map = new PingusSpotMap(level_data);
    break;
  case RANDOM:
    map = new PinguRandomMap;
    //throw PingusError("Random Map is disabled");
    break;
  default:
    throw PingusError("Error in PLF file: Undef Maptype");
    break;
  }

  if (level_data->get_mapfile().type == ResDescriptor::AUTO) {
    if (verbose) cout << "Using AutoColMap" << endl;
    colmap = map->get_colmap();
    assert(colmap);
  } else {
    if (verbose) cout << "Using StaticColMap" << endl;
    colmap = new ColMap;
    colmap->load(level_data->get_mapfile());
  }

  if (verbose)
    cout << "Loaded Colmap" << endl;

  if (verbose)
    cout << "Loaded pingu map" << endl;

  // load the background map
  bg = new Background(level_data->get_bg());

  if (verbose)
    cout << "Loaded Background" << endl;

  allowed_pingus = level_data->get_pingus();
  released_pingus = 0;
  pingus_out = 0;

  //  result->total = allowed_pingus;

  //  int pal;
  //  srand(time(NULL));

  // Creating Exit and Entrance
  vector<exit_data>     exit_d     = level_data->get_exit();
  vector<entrance_data> entrance_d = level_data->get_entrance();

  for(vector<exit_data>::size_type i10=0; i10 < exit_d.size(); ++i10) 
    exits.push_back(new Exit(exit_d[i10]));  
  
  for(vector<entrance_data>::size_type i11=0; i11 < entrance_d.size(); ++i11) 
    entrance.push_back(get_entrance(entrance_d[i11]));

  vector<trap_data>     trap_d     = level_data->get_traps();
  
  for(vector<trap_data>::size_type i12=0; i12 < trap_d.size(); ++i12)
    traps.push_back(get_trap(trap_d[i12]));

  // Creating the foreground and background hotspots
  vector<hotspot_data> hspot_d = level_data->get_hotspot();

  for(vector<hotspot_data>::size_type i = 0; i < hspot_d.size(); ++i)
    hotspot.push_back(new Hotspot(hspot_d[i]));
  
  vector<liquid_data> liquid_d = level_data->get_liquids();

  for(vector<liquid_data>::size_type i=0; i < liquid_d.size(); ++i) {
    liquid.push_back(new Liquid(liquid_d[i]));
  }

  Pingu::set_colmap(colmap);
  Pingu::set_map(map);
  Pingu::SetParticleHolder(&particle);
  Trap::SetParticleHolder(&particle);
  Entrance::SetParticleHolder(&particle);

  world_obj.push_back(map);

  // Push all objects to world_obj vector
  for(vector<hotspot_data>::size_type i = 0; i < hotspot.size(); ++i)
    world_obj.push_back(hotspot[i]);

  for(vector<exit_data>::size_type i=0; i < exits.size(); ++i)
    world_obj.push_back(exits[i]);
  
  for(vector<entrance_data>::size_type i=0; i < entrance.size(); ++i)
    world_obj.push_back(entrance[i]);

  for(vector<liquid_data>::size_type i=0; i < liquid_d.size(); ++i)
    world_obj.push_back(liquid[i]);

  for(vector<trap_data>::size_type i=0; i < traps.size(); ++i) 
    world_obj.push_back(traps[i]);

  world_obj.push_back(&pingus);

  // After all objects are in world_obj, sort them
  stable_sort(world_obj.begin(), world_obj.end(), WorldObj_less());
  
  if (verbose) {
    for(unsigned int i=0; i < world_obj.size(); ++i) 
      cout << "Z_Pos: " << world_obj[i]->z_pos << endl;
    cout << "Finished Word init" << endl;
  } 
 
  if (sound_enabled) PingusSound::play(PingusSound::LETSGO);

  // Setup the gravity force
  if (force_vectors){
    // Clear all old forces
    ForcesHolder::clear_all_forces();
    ForcesHolder::add_force(new GravityForce(grav));
  }
}

PinguHolder*
World::get_pingu_p(void)
{
  return &pingus;
}

int
World::get_width(void)
{
  assert(map);
  return map->get_width();  
}

int
World::get_height(void)
{
  assert(map);
  return map->get_height();
}

int
World::get_time(void)
{
  if (exit_world) {
    return exit_time - GameTime::get_time();
  } else {
    return GameTime::get_time();
  }
}

void
World::write_result(void)
{
  int saved   = 0;
  int killed  = 0;
  int unknown = 0;

  result.total = allowed_pingus;

  for(PinguIter pingu=pingus.begin(); pingu != pingus.end(); ++pingu) {
    switch((*pingu)->get_status()) {
    case dead:
      ++killed;
      break;
    case exited:
      ++saved;
      break;
    default:
      ++unknown;
    }
  }

  result.unknown = unknown;
  result.saved   = saved;
  result.killed  = killed;
}

Result
World::get_result()
{
  write_result();
  return result;
}

void 
World::armageddon(void)
{
  cout << "Doing armo" << endl;
  do_armageddon = true;
}

bool
World::is_finished(void)
{
  if (exit_world && exit_time < GameTime::get_time()) {
    return true;
  } else {
    return false;
  }
}

void
World::print_status(void)
{
  cout << "Released Pingus: " << released_pingus << endl;
  cout << "Allowed Pingus:  " << allowed_pingus << endl;
  cout << "Pingus Out:      " << pingus_out << endl;
}

/* EOF */
